// parseWAV.cpp
/*
 * (c)2024 Forward Computing and Control Pty. Ltd.
 * NSW Australia, www.forward.com.au
 * This code is not warranted to be fit for any purpose. You may only use it at your own risk.
 * This generated code may be freely used for both private and commercial use
 * provided this copyright is maintained.
 */


#include "parseWAV.h"
#include "SafeString.h"
#include <SD.h>

Print* parseWAV::debugPtr = NULL;

void parseWAV::setDebugOut(Print* out) {
  debugPtr = out;
}

/*
   isLittleEndian()
   returns true for little endian, false for big endian.
*/
bool parseWAV::isLittleEndian() {
  unsigned int x = 1; // atleast 16bits
  // 0x01, 0x00 if littleEndian
  // 0x00, 0x01 if bigEndian
  char *c = (char*) &x; // look at first 8bits, i.e. 0x01 little or 0x00 if big
  return ((int) * c) == 1; // cast to int will get 1 if littleEndian or 0 if bigEndian
}

// converts time duration to no of Samples (mono or stero)
size_t parseWAV::timeToSamples(pcm_wav_header_t &wf, float t_sec) {
    uint16_t channels = wf.fmt_chunk.num_of_channels;
    uint32_t freq = getSampleRate(wf);
    size_t noSamples = t_sec*freq*channels;
    return noSamples;
}


// channel numbers start from 0;
// uses current file
// returns the number of 32bit samples read, 0 if error
size_t parseWAV::read16bitData(File &file, pcm_wav_header_t &wf, int16_t *buf, size_t numOfPoints, size_t pointsOffset) {
  if (!isLittleEndian()) {
    if (debugPtr) {
      debugPtr->println("ERROR: Can run on micors that are Little Endian");
    }
    return 0;
  }
  if (!file) {
    if (debugPtr) {
      debugPtr->println("ERROR: File not open/valid");
    }
    return 0;
  }

  wav_descriptor_chunk_t *descriptor = &(wf.descriptor_chunk);
  pcm_wav_fmt_chunk_t *fmt = &(wf.fmt_chunk);
  wav_data_chunk_t *data_ck = &(wf.data_chunk);

  int numChannels = fmt->num_of_channels;
  if (numChannels > 1) {
    if (debugPtr) {
      debugPtr->println("ERROR: Can only hanle MONO (one channel)");
    }
    return 0;
  }
  int bitsPerSample = fmt->bits_per_sample;
  if (bitsPerSample != 16) {
    if (debugPtr) {
      debugPtr->println("ERROR: Can only handle 16 bit samples");
    }
    return 0;
  }

  size_t dataSize = data_ck->subchunk_size;
  size_t fileDataSize = file.size() - 44;
  if (fileDataSize < dataSize) {
    dataSize = fileDataSize;
  }

  size_t maxData = dataSize / (2); // assumes 32bit and one channel
  if (debugPtr) {
    debugPtr->print("maxData ");   debugPtr->println(maxData);
  }

  size_t offset = pointsOffset * 2; // two bytes per sample
  if (!file.seek(44 + offset)) {
    if (debugPtr) {
      debugPtr->println("Failed to get current postion");
    }
    return 0;
  }

  if (numOfPoints > (maxData - pointsOffset)) {
    if (debugPtr) {
      debugPtr->print("Asked for "); debugPtr->print(numOfPoints);
      debugPtr->print(", but only ");
      debugPtr->print(maxData - pointsOffset); debugPtr->println(" points available");
    }
    numOfPoints = maxData - pointsOffset;
  }

  size_t dataRead = (file.read((uint8_t*)buf, 2 * numOfPoints) / 2); // divide by 2 to get dataRead in samples
  if (dataRead != numOfPoints) {
    if (debugPtr) {
      debugPtr->print("Read error, read "); debugPtr->print(dataRead);
      debugPtr->print(" samples");  debugPtr->println();
    }
  }
  return dataRead;
}

void parseWAV::print16bitDataStream(File &file, pcm_wav_header_t &wf, Print * outPtr) {
  if (!outPtr) {
    return;
  }
  if (!isLittleEndian()) {
    if (outPtr) {
      outPtr->println("ERROR: Can run on micors that are Little Endian");
    }
    return;
  }
  if (!file) {
    if (debugPtr) {
      debugPtr->println("ERROR: File not open/valid");
    }
    return;
  }

  wav_descriptor_chunk_t *descriptor = &(wf.descriptor_chunk);
  pcm_wav_fmt_chunk_t *fmt = &(wf.fmt_chunk);
  wav_data_chunk_t *data_ck = &(wf.data_chunk);

  int numChannels = fmt->num_of_channels;
  if (numChannels > 1) {
    outPtr->println("ERROR: Can only hanle MONO (one channel)");
    return;
  }
  int bitsPerSample = fmt->bits_per_sample;
  if (bitsPerSample != 16) {
    outPtr->println("ERROR: Can only handle 16 bit samples");
    return;
  }
  size_t dataSize = data_ck->subchunk_size;
  size_t fileDataSize = file.size() - 44;
  if (fileDataSize < dataSize) {
    dataSize = fileDataSize;
  }

  size_t maxData = dataSize / (2); // assumes 32bit and one channel
  if (debugPtr) {
    debugPtr->print("maxData ");   debugPtr->println(maxData);
  }
  size_t dataCounter = 0;
  int32_t sample = 0;
  float deltaT = 1.0 / fmt->sample_rate;
  outPtr->println("counter,t,value");
  while (dataCounter < maxData) {
    file.read((uint8_t*)&sample, 2);
    outPtr->print(dataCounter);  outPtr->print(',');
    outPtr->print(dataCounter * deltaT, 6);  outPtr->print(',');
    outPtr->println(sample);
    dataCounter++;
  }
}



// channel numbers start from 0;
// uses current file
// returns the number of 32bit samples read, 0 if error
size_t parseWAV::read32bitData(File &file, pcm_wav_header_t &wf, int32_t *buf, size_t numOfPoints, size_t pointsOffset) {
  if (!isLittleEndian()) {
    if (debugPtr) {
      debugPtr->println("ERROR: Can run on micors that are Little Endian");
    }
    return 0;
  }
  if (!file) {
    if (debugPtr) {
      debugPtr->println("ERROR: File not open/valid");
    }
    return 0;
  }
  wav_descriptor_chunk_t *descriptor = &(wf.descriptor_chunk);
  pcm_wav_fmt_chunk_t *fmt = &(wf.fmt_chunk);
  wav_data_chunk_t *data_ck = &(wf.data_chunk);

  int numChannels = fmt->num_of_channels;
  if (numChannels > 1) {
    if (debugPtr) {
      debugPtr->println("ERROR: Can only hanle MONO (one channel)");
    }
    return 0;
  }
  int bitsPerSample = fmt->bits_per_sample;
  if (bitsPerSample != 32) {
    if (debugPtr) {
      debugPtr->println("ERROR: Can only handle 32 bit samples");
    }
    return 0;
  }

  size_t dataSize = data_ck->subchunk_size;
  size_t fileDataSize = file.size() - 44;
  if (fileDataSize < dataSize) {
    dataSize = fileDataSize;
  }

  size_t maxData = dataSize / (4); // assumes 32bit and one channel
  if (debugPtr) {
  //  debugPtr->print("maxData ");   debugPtr->println(maxData);
  }

  size_t offset = pointsOffset * 4; // two bytes per sample
  if (!file.seek(44 + offset)) {
    if (debugPtr) {
      debugPtr->println("Failed to get current postion");
    }
    return 0;
  }

  if (numOfPoints > (maxData - pointsOffset)) {
    if (debugPtr) {
      debugPtr->print("Asked for "); debugPtr->print(numOfPoints);
      debugPtr->print(", but only ");
      debugPtr->print(maxData - pointsOffset); debugPtr->println(" points available");
    }
    numOfPoints = maxData - pointsOffset;
    //return 0;
  }
  size_t dataRead = (file.read((uint8_t*)buf, 4 * numOfPoints) / 4); // divide by 4 to get dataRead in samples
  if (dataRead != numOfPoints) {
    if (debugPtr) {
      debugPtr->print("Read error, read "); debugPtr->print(dataRead);
      debugPtr->print(" samples");  debugPtr->println();
    }
  }
  return dataRead;
}

// channel numbers start from 0;
// uses current file
// pass filesize > 44 to ignore header size and use filesize
void parseWAV::print32bitDataStream(File &file, pcm_wav_header_t &wf, Print * outPtr) {
  if (!outPtr) {
    return;
  }
  if (!isLittleEndian()) {
    if (outPtr) {
      outPtr->println("ERROR: Can run only micros that are Little Endian");
    }
    return;
  }

  if (!file) {
    if (debugPtr) {
      debugPtr->println("ERROR: File not open/valid");
    }
    return;
  }

  wav_descriptor_chunk_t *descriptor = &(wf.descriptor_chunk);
  pcm_wav_fmt_chunk_t *fmt = &(wf.fmt_chunk);
  wav_data_chunk_t *data_ck = &(wf.data_chunk);

  int numChannels = fmt->num_of_channels;
  if (numChannels > 1) {
    outPtr->println("ERROR: Can only hanle MONO (one channel)");
    return;
  }
  int bitsPerSample = fmt->bits_per_sample;
  if (bitsPerSample != 32) {
    outPtr->println("ERROR: Can only handle 32 bit samples");
    return;
  }
  size_t dataSize = data_ck->subchunk_size;
  size_t fileDataSize = file.size() - 44;
  if (fileDataSize < dataSize) {
    dataSize = fileDataSize;
  }



  size_t maxData = dataSize / (4); // assumes 32bit and one channel
  if (debugPtr) {
    debugPtr->print("maxData ");   debugPtr->println(maxData);
  }
  size_t dataCounter = 0;
  int32_t sample = 0;
  float deltaT = 1.0 / fmt->sample_rate;
  outPtr->println("counter,t,value");
  while (dataCounter < maxData) {
    file.read((uint8_t*)&sample, 4);
    outPtr->print(dataCounter);  outPtr->print(',');
    outPtr->print(dataCounter * deltaT, 6);  outPtr->print(',');
    outPtr->println(sample);
    dataCounter++;
  }
}


//typedef struct {
//  char subchunk_id[4];                      /*!< Contains the letters "data" */
//  uint32_t subchunk_size;                   /*!< ==NumSamples * NumChannels * BitsPerSample / 8 */
//} __attribute__((packed)) wav_data_chunk_t; /*!< The "data" subchunk contains the size of the data and the actual sound */


//typedef struct {
//  wav_descriptor_chunk_t descriptor_chunk; /*!< Canonical WAVE format starts with the RIFF header */
//  pcm_wav_fmt_chunk_t fmt_chunk;           /*!< The "fmt " subchunk describes the sound data's format */
//  wav_data_chunk_t data_chunk;             /*!< The "data" subchunk contains the size of the data and the actual sound */
//} __attribute__((packed)) pcm_wav_header_t;

//typedef struct {
//  char subchunk_id[4];                         /*!< Contains the letters "fmt " */
//  uint32_t subchunk_size;                      /*!< PCM = 16, This is the size of the rest of the Subchunk which follows this number */
//  uint16_t audio_format;                       /*!< PCM = 1, values other than 1 indicate some form of compression */
//  uint16_t num_of_channels;                    /*!< Mono = 1, Stereo = 2, etc. */
//  uint32_t sample_rate;                        /*!< 8000, 44100, etc. */
//  uint32_t byte_rate;                          /*!< ==SampleRate * NumChannels * BitsPerSample s/ 8 */
//  uint16_t block_align;                        /*!< ==NumChannels * BitsPerSample / 8 */
//  uint16_t bits_per_sample;                    /*!< 8 bits = 8, 16 bits = 16, etc. */
//} __attribute__((packed)) pcm_wav_fmt_chunk_t; /*!< The "fmt " subchunk describes the sound data's format */

uint32_t parseWAV::getSampleRate(pcm_wav_header_t &wf) {
  return wf.fmt_chunk.sample_rate;
}
uint16_t parseWAV::getBitsPerSample(pcm_wav_header_t &wf) {
  return wf.fmt_chunk.bits_per_sample;
}
uint16_t parseWAV::getBytesPerSample(pcm_wav_header_t &wf) {
  return wf.fmt_chunk.bits_per_sample / 8;
}

uint32_t parseWAV::getNoOfChannels(pcm_wav_header_t &wf) {
  return wf.fmt_chunk.num_of_channels;
}

// need to parse first
void parseWAV::print_header(pcm_wav_header_t& wf, Print * outPtr) {
  if (!outPtr) {
    return;
  }
  wav_descriptor_chunk_t *descriptor = &(wf.descriptor_chunk);
  pcm_wav_fmt_chunk_t *fmt = &(wf.fmt_chunk);
  wav_data_chunk_t *data_ck = &(wf.data_chunk);
  cSF(sfID, 6);

  outPtr->println(" ");
  outPtr->println("Header Information");
  outPtr->println(" == == == == == == == == == ");
  outPtr->println(" ");
  sfID.clear();
  sfID.readFrom(descriptor->chunk_id, 4); // read just 4chars
  outPtr->print("  ChunkID            :  "); outPtr->println(sfID);
  outPtr->print("  ChunkSize          :  "); outPtr->println(descriptor->chunk_size);
  sfID.clear();
  sfID.readFrom(descriptor->chunk_format, 4); // read just 4chars
  outPtr->print("  Format             :  "); outPtr->println(sfID);

  sfID.clear();
  sfID.readFrom(fmt->subchunk_id, 4); // read just 4chars
  outPtr->print("  Subchunk1ID        :  "); outPtr->println(sfID);
  outPtr->print("  Subchunk1Size      :  "); outPtr->println(fmt->subchunk_size);
  outPtr->print("  AudioFormat        :  "); outPtr->println(fmt->audio_format);
  outPtr->print("  NumChannels        :  "); outPtr->println(fmt->num_of_channels);
  outPtr->print("  SampleRate         :  "); outPtr->println(fmt->sample_rate);
  outPtr->print("  ByteRate           :  "); outPtr->println(fmt->byte_rate);
  outPtr->print("  BlockAlign         :  "); outPtr->println(fmt->block_align);
  outPtr->print("  BitsPerSample      :  "); outPtr->println(fmt->bits_per_sample);

  sfID.clear();
  sfID.readFrom(data_ck->subchunk_id, 4); // read just 4chars
  outPtr->print("  Subchunk2ID        :  "); outPtr->println(sfID);
  outPtr->print("  Subchunk2Size      :  "); outPtr->println(data_ck->subchunk_size);
  outPtr->print("  Data Start         :  "); outPtr->println(44);
  size_t bytesPerSample = (size_t) - 1; // a very large number -- invalid
  if (fmt->bits_per_sample == 16) {
    bytesPerSample = 2;
    outPtr->print(" Number 16bit samples:  "); outPtr->println(data_ck->subchunk_size / bytesPerSample);
    outPtr->print("    WAV file duration:  ");   outPtr->print((float)(data_ck->subchunk_size) / bytesPerSample / fmt->num_of_channels / fmt->sample_rate);   outPtr->println(" sec ");
  } else if (fmt->bits_per_sample == 32) {
    bytesPerSample = 4;
    outPtr->print(" Number 32bit samples:  "); outPtr->println(data_ck->subchunk_size / bytesPerSample);
    outPtr->print("    WAV file duration:  ");   outPtr->print((float)(data_ck->subchunk_size) / bytesPerSample / fmt->num_of_channels / fmt->sample_rate);   outPtr->println(" sec ");
  } else {
    outPtr->print(" This BitPerSample not supported"); outPtr->println();
  }
  outPtr->println(" ");

  return;
}


/*
  note leaves file open
  updates wave_header by reference
*/
bool parseWAV::parse(File &file, pcm_wav_header_t &wf) {

  if (debugPtr) {
    debugPtr->println("Start parsing of wave file ");
  }
  if (!isLittleEndian()) {
    if (debugPtr) {
      debugPtr->println("ERROR: Can run only micros that are Little Endian");
    }
    return false;
  }

  if (!file) {
    if (debugPtr) {
      debugPtr->println("ERROR: File not open/valid");
    }
    return false;
  }

  if (!file.seek(0)) {
    // rewind file failed
    if (debugPtr) {
      debugPtr->println("Unable to rewind the file");
    }
    file.close();
    return false;
  }
  memset(&wf, 0, sizeof(wf)); // zero out header
  if (file.read((uint8_t*)(&wf), 44) != 44) {
    return false;
  }
  size_t dataSize = file.size()-44;
  // update chunck size from filesize
   wav_data_chunk_t *data_ck = &(wf.data_chunk);
   data_ck->subchunk_size = dataSize;

  return true;

}
