// writeWAV.cpp
/*
 * (c)2024 Forward Computing and Control Pty. Ltd.
 * NSW Australia, www.forward.com.au
 * This code is not warranted to be fit for any purpose. You may only use it at your own risk.
 * This generated code may be freely used for both private and commercial use
 * provided this copyright is maintained.
 */
#include "FS.h"
#include "SD.h"
#include "SPI.h"
#include "writeWAV.h"
#include "waveHeader.h"

Print* writeWAV::debugPtr = NULL;

void writeWAV::setDebugOut(Print *_debugPtr) {
  debugPtr = _debugPtr;
}

// writes with very large data size 0x40000000 1Gig bytes
// padds first sector with zero data after header if paddTo512 == true, default is false
// removes any exitsing file.
bool writeWAV::writeWaveHeader(const char *filename, uint32_t sample_width, uint32_t sample_rate, uint16_t num_channels, bool paddTo512) {
  size_t sector_size = SD.sectorSize();
  if (sector_size == 0) {
    if (debugPtr) {
      debugPtr->println("SD not started.  sectorSize() returned 0");
    }
    return false;
  }
  if ((num_channels != 1) && (num_channels != 2)) {
    if (debugPtr) {
      debugPtr->println("ERROR - num_channels must be 1 or 2, but was: "); debugPtr->println(num_channels);
    }
    return false;
  }
  uint8_t *header_buffer = (uint8_t*)calloc(sector_size, 1);
  const pcm_wav_header_t wav_header = PCM_WAV_HEADER_DEFAULT(0x40000000, sample_width, sample_rate, num_channels);
  memcpy(header_buffer, &wav_header, PCM_WAV_HEADER_SIZE);
  // reset of header_buffer is zero;
  SD.remove(filename);
  //Open the file, create one if there is no file
  File file = SD.open(filename, FILE_WRITE);
  if (!file) {
    if (debugPtr) {
      debugPtr->print("Failed to open "); debugPtr->print(filename); debugPtr->println(" for appending");
    }
    return false;
  }
  bool rtn = true;
  if (paddTo512) {
    size_t dataWritten = file.write(header_buffer, sector_size);
    if (dataWritten != sector_size) {
      if (debugPtr) {
        debugPtr->print("Write to  "); debugPtr->print(filename);  debugPtr->println(" failed");
      }
      rtn = false;
    }
  } else {
    size_t dataWritten = file.write(header_buffer, PCM_WAV_HEADER_SIZE); // only write to 44 no padding
    if (dataWritten != PCM_WAV_HEADER_SIZE) {
      if (debugPtr) {
        debugPtr->print("Write to  "); debugPtr->print(filename);  debugPtr->println(" failed");
      }
      rtn = false;
    }
  }      
  file.close();
  return rtn;
}


size_t writeWAV::write(const char *filename, uint8_t* data, size_t datasize) {
//  if (debugPtr) {
//      debugPtr->print("Appending to file: "); debugPtr->println(filename);
//  }

  File file = SD.open(filename, FILE_APPEND);
  if (!file) {
    if (debugPtr) {
      debugPtr->println("Failed to open file for appending");
    }
    return 0;
  }
  size_t dataWritten = file.write((uint8_t*) data, datasize);
  if (dataWritten != datasize) {
    if (debugPtr) {
      debugPtr->println("Append failed");
    }
    dataWritten = 0;// or return what was written??
  }
  file.close();
  return dataWritten;
}


// datasize must be multiple of sector size
size_t writeWAV::writeBlocks(const char *filename, uint8_t* data, size_t datasize) {
  size_t sector_size = SD.sectorSize();
  if (sector_size == 0) {
    if (debugPtr) {
      debugPtr->println("SD not started.  sectorSize() returned 0");
    }
    return 0;
  }
  size_t sectors_per_buffer = datasize / SD.sectorSize();
//  if (debugPtr) {
//    debugPtr->print(" sector_size:"); debugPtr->print(sector_size); debugPtr->print(" sectors per datasize:"); debugPtr->print(sectors_per_buffer);
//    debugPtr->println();
//  }
  if (datasize - (sectors_per_buffer * sector_size) != 0) {
    if (debugPtr) {
      debugPtr->print("ERROR  datasize:"); debugPtr->print(datasize); debugPtr->print(" must be multiple of "); debugPtr->println(sector_size);
    }
    return 0;
  }
//  if (debugPtr) {
//      debugPtr->print("Appending to file: "); debugPtr->println(filename);
//  }

  File file = SD.open(filename, FILE_APPEND);
  if (!file) {
    if (debugPtr) {
      debugPtr->println("Failed to open file for appending");
    }
    return 0;
  }
  size_t dataWritten = file.write((uint8_t*) data, datasize);
  if (dataWritten != datasize) {
    if (debugPtr) {
      debugPtr->println("Append failed");
    }
    dataWritten = 0;// or return what was written??
  }
  file.close();
  return dataWritten;
}
